# -*- coding: utf-8 -*-

"""Species and gene tree visualization."""

import numpy as np
import matplotlib.pyplot as plt
import matplotlib.patches as mpatches


__author__ = 'David Schaller'


def assign_species_colors(tree):
    """Assign a unique color to each (extant) species in a species tree.
    
    Parameters
    ----------
    tree : Tree
        The species tree with uniquely labeled (non-loss) leaves.
    
    Returns
    -------
    dict
        A dictonary containing the labels of the extant species as keys and
        the assigned colors as values.
    """
    
    return _assign_cmap({v.label for v in tree.leaves() if v.event != 'L'})


def assign_gene_colors(tree, species_colors=None):
    """Assign a color to each (extant) gene in a gene tree according to its
    species.
    
    Parameters
    ----------
    tree : Tree
        The gene tree whose (non-loss) leaves have the 'reconc' attribute, i.e,
        the information to which species they belong.
    species_colors : dict, optional
        A dictonary containing the values of the 'color' attribute appearing
        in the tree as keys and assigned colors as values.
    
    Returns
    -------
    dict
        A dictonary containing the labels of the extant genes as keys and
        the assigned colors as values.
    """
    
    if species_colors is None:
        
        species_colors = _assign_cmap({v.reconc for v in tree.leaves()
                                       if v.event != 'L'})
    
    return {v.label: species_colors[v.reconc] for v in tree.leaves()
            if v.event != 'L'}


def assign_colors(species_tree, gene_tree):
    """Assign a color to each (extant) species and gene.
    
    Parameters
    ----------
    species_tree : Tree
        The species tree with uniquely labeled (non-loss) leaves.
    gene_tree : Tree
        The gene tree whose (non-loss) leaves have the 'reconc' attribute, i.e,
        the information to which species they belong, that furthermore appear
        as labels of the (non-loss) leaves in the species tree.
    
    Returns
    -------
    tuple of two dicts
        Dictonaries containing the labels of the extant species/genes as keys
        and the assigned colors as values.
    """
    
    species_colors = assign_species_colors(species_tree)
    
    return species_colors, assign_gene_colors(gene_tree,
                                              species_colors=species_colors)


def _assign_cmap(colors):
    
    color_dict = {}
    
    if len(colors) <= 10:
        cmap = plt.get_cmap('tab10')(np.arange(len(colors), dtype=int))
    else:
        cmap = plt.get_cmap('jet')(np.linspace(0, 1.0, len(colors)))
        
    for label, color in zip(colors, cmap):
        color_dict[label] = color
    
    return color_dict


def visualize(tree, color_dict=None, save_as=False, scale_symbols=1.0,
              fontsize='medium'):
    """Visualize a tree.
    
    Parameters
    ----------
    tree : Tree
        A tree whose nodes have the 'dist' attribute.
    color_dict : dict, optional
        A dictonary containing as values the assigned color for each label
        of the (non-loss) leaves.
    save_as : str, optional
        Save the image to a file. The default is False, in which case the
        image is not saved.
    scale_symbols : float, optional
        Adjust the size of the event type symbols. The default is 1.0.
    fontsize : str or float or int, optional
        Adjust the size of the text. The default is 'medium'.
    """
    
    vis = Visualizer(tree, color_dict=color_dict, scale_symbols=scale_symbols,
                     fontsize=fontsize)
    vis.draw(save_as=save_as)


class Visualizer:
    """Species and gene tree visualization."""
    
    def __init__(self, tree, color_dict=None,
                 scale_symbols=1.0, fontsize='medium',
                 species_info=False):
        """
        Parameters
        ----------
        tree : Tree
            A tree whose nodes have the 'dist' attribute.
        color_dict : dict, optional
            A dictonary containing as values the assigned color for each label
            of the (non-loss) leaves.
        scale_symbols : float, optional
            Adjust the size of the event type symbols. The default is 1.0.
        fontsize : str or float or int, optional
            Adjust the size of the text. The default is 'medium'.
        species_info : bool, False
            If True, write the reconciliation information behind the leaf
            label if available. The default is False.
        """
        
        self.tree = tree
        
        if color_dict:
            self.colors = color_dict
        else:
            self.colors = {v.label: 'white' for v in tree.preorder()}
        
        self.species_info = species_info
        
        self.symbolsize = 0.03 * scale_symbols
        self.fontsize = fontsize
        self.symbollw = 0.04
        self.leafs_per_vertical_unit = 15
        self.symbol_zorder = 3
        
        self.distance_dict = {}
        #self.divtime_dict {}
        self.leaf_counter = sum(1 for _ in tree.leaves())
        self.node_positions = {}
        
    
    def draw(self, distance_mode='evolutionary', save_as=False):
        """Draw the tree and optionally save it to file.
        
        Parameters
        ----------
        distance_mode : str, optional
            Information that is used for the length of the species edges.
            Currently only 'evolutionary' is supported, in which case the
            'dist' attribute is used.
        save_as : str, optional
            Save the image to a file. The default is False, in which case the
            image is not saved.
        """
        
        self.fig, self.ax = plt.subplots()
        self.ax.set_aspect('equal')
        self.ax.invert_yaxis()
        
        self.initial_traversal(distance_mode)
        # self.assign_colors()
        self.assign_positions()
        self.draw_edges()
        self.draw_nodes()
        
        self.ax.axvline(x=0, linewidth=1, color='grey', linestyle='--')
        self.ax.set_yticks([])
        self.ax.spines['top'].set_visible(False)
        self.ax.spines['left'].set_visible(False)
        self.ax.spines['right'].set_visible(False)
        self.ax.tick_params(axis='x', labelsize=self.fontsize)
        
        xmin, xmax = self.ax.get_xlim()
        ymin, ymax = self.ax.get_ylim()
        self.fig.set_size_inches(5*abs(xmax-xmin), 5*abs(ymax-ymin)+0.4)
        plt.tight_layout()
        if save_as:
            plt.savefig(save_as, dpi=300)
        else:
            plt.show()
        
    
    def initial_traversal(self, distance_mode):
        
        xmax = 0.0
        
        for v in self.tree.preorder():
            if distance_mode == 'evolutionary':
                if not v.parent:
                    self.distance_dict[v] = 0.0
                else:
                    self.distance_dict[v] = self.distance_dict[v.parent] + v.dist
                    if self.distance_dict[v] > xmax:
                        xmax = self.distance_dict[v]
            elif distance_mode == 'divergence time':
                raise ValueError('divergence time not yet implemented')
            else:
                raise ValueError(f"distance mode '{distance_mode}' not supported")
                
        self.ax.set_xlim(-0.1,xmax+0.5)
    
    
    def assign_positions(self):
        
        ymax = (self.leaf_counter-1)/self.leafs_per_vertical_unit
        self.ax.set_ylim(ymax+0.1, -self.symbolsize*0.6)
        
        yposition = 0
        for v in self.tree.postorder():
            if not v.children:
                self.node_positions[v] = (self.distance_dict[v],
                                          yposition)
                yposition += 1/self.leafs_per_vertical_unit
            else:
                ymean = (self.node_positions[v.children[0]][1] +
                         self.node_positions[v.children[-1]][1])/2
                self.node_positions[v] = (self.distance_dict[v],
                                          ymean)
    
    
    def draw_edges(self):
        
        for v in self.tree.preorder():
            if v.parent:
                if hasattr(v, 'transferred') and v.transferred:
                    color = 'red'
                else:
                    color = 'black'
                self.ax.plot([self.node_positions[v.parent][0],
                              self.node_positions[v][0]],
                             [self.node_positions[v][1],
                              self.node_positions[v][1]],
                     color=color,
                     linestyle='-', linewidth=1)
            if v.children:
                self.ax.plot([self.node_positions[v][0],
                              self.node_positions[v][0]],
                             [self.node_positions[v.children[0]][1],
                              self.node_positions[v.children[-1]][1]],
                     color='black',
                     linestyle='-', linewidth=1)
    
    
    def draw_nodes(self):
        
        for v in self.tree.preorder():
            if not v.parent and not v.event:
                self.draw_root(*self.node_positions[v])
            elif v.children:
                if v.event == 'D':
                    self.draw_dupl(*self.node_positions[v])
                elif v.event == 'S':
                    self.draw_spec(*self.node_positions[v])
                elif v.event == 'H':
                    self.draw_hgt(*self.node_positions[v])
                elif v.event == 'GC':
                    self.draw_gc(*self.node_positions[v])
            else:
                if v.event == 'L':
                    self.draw_loss(*self.node_positions[v])
                else:
                    x, y = self.node_positions[v]
                    self.draw_leaf(x, y, color=self.colors[v.label])
                    if not self.species_info:
                        text = str(v.label)
                    elif hasattr(v, 'reconc'):
                        text = '{} <{}>'.format(v.label, v.reconc)
                    self.write_label(x+self.symbolsize+0.02, y, text)
                
    
    def draw_leaf(self, x, y, color='white', leftalign=False):
        
        if leftalign:
            x += self.symbolsize/2
        fill = mpatches.Circle((x, y), self.symbolsize/2,
                               color=color, fill=True,
                               zorder=self.symbol_zorder)
        self.ax.add_patch(fill)
        outer = mpatches.Circle((x, y), self.symbolsize/2,
                                color='black', fill=False,
                                lw=self.symbolsize/self.symbollw,
                                zorder=self.symbol_zorder)
        self.ax.add_patch(outer)
        little = mpatches.Circle((x, y), self.symbolsize/8,
                                 color='black', fill=True,
                                 zorder=self.symbol_zorder)
        self.ax.add_patch(little)
        
    
    def draw_loss(self, x, y):
        
        self.ax.plot([x,x],
                     [y-self.symbolsize/2, y+self.symbolsize/2],
                     color='black',
                     linestyle='-', linewidth=1)
    
    
    def draw_root(self, x, y, rightalign=False):
        
        if rightalign:
            x -= self.symbolsize/2
        fill = mpatches.Circle((x, y), self.symbolsize/2,
                               color='white', fill=True,
                               zorder=self.symbol_zorder)
        self.ax.add_patch(fill)
        outer = mpatches.Circle((x, y), self.symbolsize/2,
                                color='black', fill=False,
                                lw=self.symbolsize/self.symbollw,
                                zorder=self.symbol_zorder)
        self.ax.add_patch(outer)
        little = mpatches.Circle((x, y), self.symbolsize/5,
                                 color='black', fill=False,
                                 lw=self.symbolsize/self.symbollw,
                                 zorder=self.symbol_zorder)
        self.ax.add_patch(little)
    
    
    def draw_spec(self, x, y):
        
        fill = mpatches.Circle((x, y), self.symbolsize/2,
                               color='black', fill=True,
                               zorder=self.symbol_zorder)
        self.ax.add_patch(fill)
        
    
    def draw_dupl(self, x, y):
        
        square = mpatches.Rectangle((x-self.symbolsize/2,
                                     y-self.symbolsize/2),
                                    width=self.symbolsize,
                                    height=self.symbolsize,
                                    color='white', fill=True,
                                    zorder=self.symbol_zorder)
        self.ax.add_patch(square)
        border = mpatches.Rectangle((x-self.symbolsize/2,
                                     y-self.symbolsize/2),
                                    width=self.symbolsize,
                                    height=self.symbolsize,
                                    color='black', fill=False,
                                    lw=self.symbolsize/self.symbollw,
                                    zorder=self.symbol_zorder)
        self.ax.add_patch(border)
    
    
    def draw_hgt(self, x, y):
        
        coord = np.asarray([[x-self.symbolsize/2,y],
                            [x+self.symbolsize/2,y-self.symbolsize/2],
                            [x+self.symbolsize/2,y+self.symbolsize/2]])

        inner = mpatches.Polygon(coord, closed=True,
                                 color='white', fill=True,
                                 zorder=self.symbol_zorder)
        self.ax.add_patch(inner)
        outer = mpatches.Polygon(coord, closed=True,
                                 color='black', fill=False,
                                 lw=self.symbolsize/self.symbollw,
                                 zorder=self.symbol_zorder)
        self.ax.add_patch(outer)
    
    
    def draw_gc(self, x, y):
        
        coord = []
        R = self.symbolsize/2
        for i in range(5):
            coord.append([x + 0.5 * R * np.cos(np.radians(0+i*72)),
                          y - 0.5 * R * np.sin(np.radians(0+i*72))])
            coord.append([x + R * np.cos(np.radians(36+i*72)),
                          y - R * np.sin(np.radians(36+i*72))])

        inner = mpatches.Polygon(coord, closed=True,
                                 color='black', fill=True,
                                 zorder=self.symbol_zorder)
        self.ax.add_patch(inner)
    
    
    def write_label(self, x, y, text):
        
        self.ax.text(x, y, text,
                     fontsize=self.fontsize,
                     horizontalalignment='left',
                     verticalalignment='center')



# -*- coding: utf-8 -*-

"""Substitution models for nucleotide and amino acid evolution.

References
----------
.. [1] Z. Yang.
   Computational molecular evolution.
   Oxford series in ecology and evolution. Oxford University Press, 2006.
   ISBN 978-0-19-856699-1 978-0-19-856702-8.
"""

import numpy as np
from scipy import linalg

from asymmetree.seqevolve.EmpiricalModels import empirical_models
from asymmetree.file_io.SubstModelIO import parse_paml


__author__ = 'David Schaller'


class SubstModel:
    """Substitution model for nucleotide or amino acid evolution."""
    
    nuc_models = {'JC69', 'K80', 'GTR', 'CUSTOM'}
    aa_models = {'JC69', 'CUSTOM'}
    
    nucleotides = 'ACGT'
    amino_acids = 'ARNDCQEGHILKMFPSTWYV'
    
    
    def __init__(self, model_type, model_name,
                 **kwargs):
        """
        Parameters
        ----------
        model_type : str
            Available options are 'n' for nucleotide and 'a' for amino acid
            sequences.
        model_name : str
            Availbale are e.g. 'JC69' (nuc./a.a.), 'K80' (nuc.), 'GTR' (nuc.),
            'WAG' (a.a.), 'JTT' (a.a.), 'BLOSUM62' (a.a.), 'LG' (a.a.), and
            'DAYHOFF' (a.a.).
            For the option 'CUSTOM', the keyword parameter 'filename' has to
            be specified with a path to a custom model in paml format.
        kwargs : optional
            Some models require additional parameters, e.g. 'K80' requires the
            parameter 'kappa' and 'GTR' requires 'abcdef' and 'f', see
            documention for details.
        
        Raises
        ------
        ValueError
            If an invalid model was specified.
        """
        
        self.model_type = model_type.lower()
        self.model_name = model_name.upper()
        
        self._params = kwargs
        
        if (self.model_type in ('n', 'nuc', 'nucleotide') and
            self.model_name in SubstModel.nuc_models):
            
            self.model_type = 'n'
            self.alphabet = SubstModel.nucleotides
            
        elif (self.model_type in ('a', 'aa', 'amino', 'aminoacid', 'protein') and
              (self.model_name in SubstModel.aa_models or 
               self.model_name in empirical_models)):
            
            self.model_type = 'a'
            self.alphabet = SubstModel.amino_acids
                
        else:
            raise ValueError("model '{}', '{}' is not "\
                             "available".format(model_type, model_name))
        
        self.alphabet_dict = {item: index for index, item in enumerate(self.alphabet)}
        
        self._load_exchangeability_and_freqs()
        self._build_rate_matrix()
        
    
    def _load_exchangeability_and_freqs(self):
        """Load the exchangeability matrix S and the stationary frequencies pi."""
        
        # a custom model (via a paml file) was specified
        if self.model_name == 'CUSTOM':
            
            if self._params is None or 'filename' not in self._params:
                    raise ValueError("custom model requires the parameter "\
                                     "'filename'")
            
            S, freqs = parse_paml(self._params['filename'],
                                  model_type=self.model_type)
            self.S, self.freqs = np.asarray(S), np.asarray(freqs)
        
        # non-empirical nucleotide models
        elif self.model_type == 'n':
            
            if self.model_name == 'JC69':
                self.S, self.freqs = _JC69_nuc()
                
            elif self.model_name == 'K80':
                
                if self._params is None or 'kappa' not in self._params:
                    raise ValueError("model 'K80' requires the parameter 'kappa'")
                
                self.S, self.freqs = _K80_nuc(self._params['kappa'])
            
            elif self.model_name == 'GTR':
                
                if self._params is None or 'abcdef' not in self._params or 'f' not in self._params:
                    raise ValueError("model 'GTR' requires the parameters "\
                                     "'abcdef' and 'f'")
                
                self.S, self.freqs = _GTR_nuc(self._params['abcdef'],
                                              self._params['f'])
        
        # non-empirical and empirical amino acid models
        elif self.model_type == 'a':
            
            if self.model_name == 'JC69':
                self.S, self.freqs = _JC69_aa()
                
            elif self.model_name in empirical_models:
                self.S, self.freqs = empirical_models[self.model_name]()
                
        # make sure stationary frequencies sum to 1
        self.freqs /= np.sum(self.freqs)
        
        # compute cumulative frequencies for evolver
        self.freqs_cumulative = np.cumsum(self.freqs)
    
    
    def _build_rate_matrix(self):
        """Compute the rate matrix Q."""
    
        # rate matrix from exchangeability matrix and stationary frequencies
        Q = self.S @ np.diag(self.freqs)
        
        # compute diagonals, rows sum to 0
        for i in range(Q.shape[0]):
            Q[i, i] = - (np.sum(Q[i, :]) - Q[i, i])
            
        # normalize matrix
        k = - np.dot(self.freqs, np.diagonal(Q))
        Q /= k
        
        self.Q = Q
    
    
    def eigensystem(self):
        
        if not hasattr(self, 'eigenvals'):
            self.eigenvals, self.U, self.U_inv = diagonalize(self.Q, self.freqs)
        
        return self.eigenvals, self.U, self.U_inv
    
    
    def transition_prob_matrix(self, t):
        """Calculate the transition probability matrix P(t).
        
        P(t) = U x e^(Lambda*t) x U^(-1).
        
        Parameters
        ----------
        t : float
            The time / evolutinary distance.
        
        Returns
        -------
        2-dimensional numpy array
        """
        
        # ensure that eigensystem has been computed
        self.eigensystem()
        
        # first multiplication element-wise, since corresponding matrix
        # only has non-zero entries on the diagonal
        
        return (self.U * np.exp(self.eigenvals * t))  @  self.U_inv
        
    
    def to_indices(self, sequence):
        """Convert a sequence into a list of indices the represent the positions
        of the letters in the alphabet of the model.
        
        Parameters
        ----------
        sequence : iterable
            The sequence to be converted.
        
        Returns
        -------
        list
            The list of indices.
        
        Raises
        ------
        ValueError
            If the sequence contains a letter that is not in the alphabet of
            the model.
        """
        
        try:
            result = [self.alphabet_dict[letter] for letter in sequence]
        except KeyError:
            raise ValueError('invalid sequence for the specified model')
            
        return result
    
    
    def to_sequence(self, evoseq):
        """Construct the str representation of a sequence.
        
        Parameters
        ----------
        evoseq : asymmetree.seqevolve.EvolvingSequence.EvoSeq
            The sequence to be converted.
        
        Returns
        -------
        str
            The nucleotide or amino acid sequence depending on the model.
        """
        
        return ''.join(self.alphabet[x._value] for x in evoseq)
    

# --------------------------------------------------------------------------
#                           Module functions
# --------------------------------------------------------------------------
    
    
def diagonalize(Q, freqs):
    """Diagonalize a matrix.
    
    Parameters
    ----------
    Q : 2-dimensional numpy array
        Rate matrix.
    freqs : 1-dimensional numpy array
        stationary frequencies
    
    Returns
    -------
    1-dimensional numpy array
        Eigenvalues.
    2-dimensional numpy array
        Diagonalized matrix.
    2-dimensional numpy array
        Inverse of the diagonalized matrix.
    """
        
    # matrix is already symmetric
    if np.allclose(Q, Q.T):
        
        eigenvals, U = linalg.eigh(Q)
        U_inv = linalg.inv(U)
    
    # matrix is not symmetric but model is time-reversible
    else:
    
        Phi = np.diag(np.sqrt(freqs))
        Phi_inv = linalg.inv(Phi)
        
        B = Phi @ Q @ Phi_inv
        
        eigenvals, R = linalg.eigh(B)

        U = Phi_inv @ R
        U_inv = linalg.inv(R) @ Phi
    
    return eigenvals, U, U_inv
    
    
# --------------------------------------------------------------------------
#                           NUCLEOTIDE MODELS
# --------------------------------------------------------------------------

def _JC69_nuc():
    
    S = np.array([[0, 1, 1, 1],
                  [1, 0, 1, 1],
                  [1, 1, 0, 1],
                  [1, 1, 1, 0]])
    
    freqs = np.array([0.25, 0.25, 0.25, 0.25])
    
    return S, freqs


def _K80_nuc(kappa):
    
    S = np.array([[0,     1,     kappa, 1    ],
                  [1,     0,     1,     kappa],
                  [kappa, 1,     0,     1    ],
                  [1,     kappa, 1,     0    ]])
    
    freqs = np.array([0.25, 0.25, 0.25, 0.25])
    
    return S, freqs


def _GTR_nuc(abcdef, f):
    """Generalized time-reversible model.
    
    Parameterization as in e.g. used in PAML and ALF:
    a:    C <--> T
    b:    A <--> T
    c:    G <--> T
    d:    A <--> C  
    e:    C <--> G
    f:    A <--> G
    """
    
    a, b, c, d, e, f = abcdef
    
    S = np.array([[0,  d,  f,  b],
                  [d,  0,  e,  a],
                  [f,  e,  0,  c],
                  [b,  a,  c,  0]])
    
    freqs = np.asarray(abcdef)
    freqs /= np.sum(freqs)
    
    return S, freqs


# --------------------------------------------------------------------------
#                          AMINO ACID MODELS
# --------------------------------------------------------------------------

def _JC69_aa():
    
    S = np.ones((20, 20))
    S -= np.identity(20)
    
    freqs = np.full((20,), 1/20)
    
    return S, freqs


# -*- coding: utf-8 -*-

"""Extraction and analysis of orthology, xenology, best matches, ...
"""

from asymmetree.analysis.BestMatches import (orthology_from_tree,
                                             bmg_from_tree,
                                             extended_best_hits,
                                             is_bmg,
                                             lrt_from_tree,
                                             lrt_from_colored_graph,
                                             binary_refinable_tree,
                                             augment_and_label,)

from asymmetree.analysis.HGT import (true_transfer_edges,
                                     rs_transfer_edges,
                                     fitch,
                                     undirected_fitch,
                                     is_rs_fitch,
                                     below_equal_above,
                                     ldt_graph,
                                     RsScenarioConstructor)



# -*- coding: utf-8 -*-

from tralda.datastructures import TreeNode


__author__ = 'David Schaller'


# ------------------------------------------------------------------------------
#                          Auxiliary functions
# ------------------------------------------------------------------------------

def labeled_sequences(sequences):
    """Converts dictionaries to lists and keys into strings.
    
    If the keys are 'TreeNode' instance, then their label is taken. Also works
    if 'sequences' already is a list (of tuples).
    """
    
    result = []
    
    if isinstance(sequences, dict):
        sequences = sequences.items()
    
    for key, sequence in sequences:
        
        if isinstance(key, TreeNode):
            result.append( (str(key.label), sequence) )
        else:
            result.append( (str(key),  sequence) )
            
    return result


# ------------------------------------------------------------------------------
#                               Fasta files
# ------------------------------------------------------------------------------

def write_fasta(filename, sequences):
    
    sequences = labeled_sequences(sequences)
    
    with open(filename, 'w') as f:
        
        for label, seq in sequences:
            
            f.write('>{}\n'.format(label))
            pos = 0
            while pos < len(seq):
                f.write(seq[pos:min(pos+80, len(seq))])
                pos += 80
                f.write('\n')


# ------------------------------------------------------------------------------
#                       Multiple Sequence Alignments
# ------------------------------------------------------------------------------

def write_alignment(filename, alignment, alignment_format='phylip'):
    
    alignment = labeled_sequences(alignment)
    
    with open(filename, 'w') as f:
        
        if alignment_format == 'phylip':
            _write_phylip(f, alignment)
    
        elif alignment_format == 'clustal':
            _write_clustal(f, alignment)
        
        elif alignment_format == 'pretty':
            _write_pretty(f, alignment)
            
        else:
            raise ValueError("alignment format '{}' is not "\
                             "available".format(alignment_format))
            

def _check_alignment(alignment):
    
    max_length = 0                          # maximal label length
    seq_length = None                       # length of the aligned sequences
    for label, seq in alignment:
        
        if len(str(label)) > max_length:
            max_length = len(str(label))
            
        if seq_length is None:
            seq_length = len(seq)
        elif seq_length != len(seq):
            raise ValueError('aligned sequences must have the same length')
            
    return max_length, seq_length
            
            
def _write_phylip(f, alignment):
    
    max_length, seq_length = _check_alignment(alignment)
    max_length = max(max_length, 9)
    
    
    f.write('  {} {} i'.format(len(alignment), seq_length))
    
    format_str = '\n{:<' + str(max_length+1) + '}'
    current = 0
    
    while current < seq_length:
        
        end = min(seq_length, current+50)
        
        for label, seq in alignment:
            
            if current == 0:
                f.write(format_str.format(label))
            else:
                f.write(format_str.format(''))
                
            f.write(' '.join( seq[i:min(i+10,end)] for i in range(current, end, 10) ))
                
        if end != seq_length:
            f.write('\n')
        
        current += 50


def _write_clustal(f, alignment):
    
    max_length, seq_length = _check_alignment(alignment)
    
    f.write('CLUSTAL W (1.8) multiple sequence alignment\n\n')
    
    format_str = '\n{:<' + str(max_length+4) + '}{}'
    current = 0
    
    while current < seq_length:
        
        end = min(seq_length, current+60)
        
        for label, seq in alignment:
            f.write(format_str.format(label, seq[current:end]))
                
        if end != seq_length:
            f.write('\n\n')
        
        current += 60
    
        
def _write_pretty(f, alignment):
    
    _, seq_length = _check_alignment(alignment)
    
    f.write('  1          11         21         31         41       50\n')
    f.write('  |          |          |          |          |        |')
    
    current = 0
    
    while current < seq_length:
        
        end = min(seq_length, current+50)
        
        for label, seq in alignment:
            
            count = end - current - seq[current:end].count('-')
            seq_string = ' '.join( seq[i:min(i+10,end)] for i in range(current, end, 10) )
            f.write('\n  {:54}{:>6} {}'.format(seq_string, count, label))
                
        if end != seq_length:
            f.write('\n\n\n')
        
        current += 50


# -*- coding: utf-8 -*-

"""Simulation of genomes."""

import os

from tralda.datastructures.Tree import Tree

import asymmetree.treeevolve as te
import asymmetree.seqevolve as se
from asymmetree.file_io.SeqFileIO import write_alignment, write_fasta
from asymmetree.tools.Sampling import Sampler


__author__ = 'David Schaller'


class GenomeSimulator:
    
    def __init__(self, species_tree, outdir=None):
        """
        Parameters
        ----------
        species_tree : Tree
            The species tree along which gene trees are simulated.
        outdir : str, optional
            The path to a directory into which the results of the simulation
            (serialized trees, one fasta file per genome, one 'true' alignement
            per gene family). The default is None, in which case nothing is
            saved to file.
        """
        
        if not isinstance(species_tree, Tree):
            raise TypeError("species tree must be of type 'Tree'")
        
        self.S = species_tree
        self.outdir = outdir
        
        if self.outdir:
            self._check_outdir()
            self.S.serialize(self._path('species_tree.json'), mode='json')
            
        self.true_gene_trees = []
        self.pruned_gene_trees = []
        
    
    def _check_outdir(self):
        
        if not os.path.exists(self.outdir):
            os.makedirs(self.outdir)
            
        elif os.path.exists(self.outdir) and not os.path.isdir(self.outdir):
            raise FileExistsError("'{}' is not a directory".format(self.outdir))
            
        for directory in ('true_gene_trees', 'fasta_files', 'alignments'):
            path = os.path.join(self.outdir, directory)
            if not os.path.exists(path):
                os.makedirs(path)
            
    
    def _path(self, *args):
        
        return os.path.join(self.outdir, *args)
            
    
    def simulate_gene_trees(self, n,
                            **kwargs):
        """Simulate gene trees.
        
        Simulates the 'true gene trees' which still contain the loss events as
        well as the 'pruned gene trees' in which the loss branches are
        removed.
        
        Parameters
        ----------
        n : int
            Number of gene trees to be simulated.
        kwargs : optional
            See function 'gene_trees' in 'treeevolve' subpackage or 
            documentation.
        """
        
        self.number_of_families = n
        
        self.true_gene_trees = te.gene_trees(self.S, n=n, **kwargs)
        
        self.pruned_gene_trees = [te.prune_losses(tree)
                                  for tree in self.true_gene_trees]
        
        # sequences should be emptied here if methods were called before
        if hasattr(self, 'sequence_dicts'):
            self.sequence_dicts.clear()
        
        if self.outdir:
            for i in range(n):
                filename = self._path('true_gene_trees',
                                      'gene_tree{}.json'.format(i))
                self.true_gene_trees[i].serialize(filename, mode='json')
        
        return self.true_gene_trees, self.pruned_gene_trees
                
    
    def simulate_sequences(self, subst_model,
                           root_genome=None,
                           length_distr=('constant', 200),
                           min_length=10,
                           max_length=None,
                           write_fastas=True,
                           write_alignments=True,
                           **kwargs):
        """Simulate sequences along the (pruned) gene trees.
        
        Parameters
        ----------
        subst_model : SubstModel
            Substitution model.
        root_genome : list, optional
            List of sequences for the roots of the gene trees, must contain the
            same number of str sequences as trees that were simulated. 
            The sequences must be compatible with the specified substitution 
            model (model_type='n'/'a'). The default is None, in which case new
            sequences are generated at random.
        length_distr : tuple, optional
            Distribution of the length of the root sequences if root genome is
            not supplied. See documentation for available options.
        min_length : int, optional
            Minimal length at which the distribution of lengths is truncated;
            must be less than the mean of this distribution. The default is 10.
        max_length : int, optional
            Maximal length at which the distribution of lengths is truncated;
            must be less than the mean of this distribution. The default is
            None, in which case the distribution is not truncated.
        write_fastas : bool, optional
            If True and an output directory was speci ed, write the sequences
            (one file per species) into the directory 'fasta_files' in the
            output directory. The default is True.
        write_alignments : bool, optional
            If True and an output directory was speci ed, write the true 
            alignments (one file per gene tree) into the directory 'alignments'
            in the output directory. The default is True.
        kwargs : optional
            See class 'Evolver' in the 'seqevolve' subpackage or documentation.
        """
        
        self.subst_model = subst_model
        
        if hasattr(self, 'sequence_dicts'):
            self.sequence_dicts.clear()
        else:
            self.sequence_dicts = []
        
        if root_genome:
            if len(root_genome) != self.number_of_families:
                raise ValueError('no. of sequences in root genome does not'\
                                 'match no of gene families')
        else:
            if isinstance(length_distr, Sampler):
                self.sampler = length_distr
            else:
                self.sampler = Sampler(length_distr,
                                       minimum=min_length,
                                       maximum=max_length,
                                       discrete=True)
        
        evolver = se.Evolver(subst_model, **kwargs)
        
        for i in range(self.number_of_families):
            
            PGT = self.pruned_gene_trees[i]
            
            if root_genome:
                evolver.evolve_along_tree(PGT, start_seq=root_genome[i].upper())
                
            else:
                evolver.evolve_along_tree(PGT, start_length=self.sampler())
                
            self.sequence_dicts.append(evolver.sequences)
            
            # write one alignment file per gene family
            if self.outdir and write_alignments:
                self._write_alignment(i)
        
        # write one fasta file per species
        if self.outdir and write_fastas:
            self._write_fastas(include_inner=False)
            
            
    def _write_alignment(self, family_id):
        
        if not self.outdir:
            raise RuntimeError('no output directory specified for alignments')        
        
        alg_builder = se.AlignmentBuilder(self.pruned_gene_trees[family_id],
                                          self.sequence_dicts[family_id],
                                          self.subst_model.alphabet,
                                          include_inner=False)
        
        alignment = []
        for node, sequence in alg_builder.build().items():
            label = self._compose_label(node, family_id)
            alignment.append( (label, sequence) )
            
        basename = 'alignment{}.phylip'.format(family_id)
        filename = self._path('alignments', basename)
        write_alignment(filename, alignment, alignment_format='phylip')
    
                
    def _write_fastas(self, include_inner=False):
        
        if not self.outdir:
            raise RuntimeError('no output directory specified for fasta files')
        
        # list of all species labels
        if include_inner:
            species = [v.label for v in self.S.preorder()]
        else:
            species = [v.label for v in self.S.preorder() if not v.children]
        
        # labeled sequences sorted by species
        sorted_seqs = {s: [] for s in species}
        
        for i in range(self.number_of_families):
            for node, evoseq in self.sequence_dicts[i].items():
                
                # skip inner nodes
                if not include_inner and node.children:
                    continue
                
                label = self._compose_label(node, i)
                sequence = self.subst_model.to_sequence(evoseq)
                sorted_seqs[node.reconc].append( (label, sequence) )
                
        for spec, sequences in sorted_seqs.items():
            basename = '{}.f{}a'.format(spec, self.subst_model.model_type)
            filename = self._path('fasta_files', basename)
            write_fasta(filename, sequences)
        
    
    def _compose_label(self, node, family_id):
        
        return 'fam{}gene{}spec{}'.format(family_id, node.label, node.reconc)



# -*- coding: utf-8 -*-

"""Evolving sequences along trees.

The subpackage asymmetree.seqevolve contains modules for the simulation of
nucleotide or amino acid sequences along a phylogenetic tree using
time-continuous Markov chains, as usually applied for this purpose (for
textbooks, see e.g. [1], [2], and [3]).

These models typically take a substitution-rate matrix and the equilibrium
frequencies of the states (i.e. the nucleotides or amino acids as input).
Moreover, insertions and deletions (indels) and heterogeneity among the sites 
can be simulated.

References
----------
.. [1] J. Felsenstein.
   Inferring Phylogenies. Sinauer Associates, Sunderland, Mass, 2004.
   ISBN 978-0-87893-177-4.
.. [2] Z. Yang.
   Computational molecular evolution.
   Oxford series in ecology and evolution. Oxford University Press, 2006.
   ISBN 978-0-19-856699-1 978-0-19-856702-8.
.. [3] Z. Yang.
   Molecular Evolution: A Statistical Approach.
   Oxford University Press, Oxford, United Kingdom; New York, NY, United
   States of America,  rst edition edition, 2014.
   ISBN 978-0-19-960261-2 978-0-19-960260-5.
"""

from asymmetree.seqevolve.Evolver import Evolver
from asymmetree.seqevolve.SubstModel import SubstModel
from asymmetree.seqevolve.IndelModel import IndelModel
from asymmetree.seqevolve.HetModel import HetModel
from asymmetree.seqevolve.Alignment import AlignmentBuilder
from asymmetree.seqevolve.EvolvingSequence import EvoSeq


# -*- coding: utf-8 -*-

"""
Simulation of dated gene trees.
"""

import random, warnings
from collections import deque
from dataclasses import dataclass
from warnings import warn

import numpy as np

from tralda.datastructures.Tree import Tree, TreeNode, LCA

from asymmetree.tools.PhyloTreeTools import (sorted_nodes,
                                             sorted_edges,
                                             delete_losses_and_contract,
                                             remove_planted_root,
                                             distance_from_timing)


__author__ = 'David Schaller'


# --------------------------------------------------------------------------
#                    CONSTRUCTION OF THE GENE TREE
# --------------------------------------------------------------------------
    

def dated_gene_tree(S, **kwargs):
    """Simulate a gene tree along the specified species tree.
    
    Parameters
    ----------
    dupl_rate : float, optional
        Duplication rate, default is 0.0.
    loss_rate : float, optional
        Loss rate, default is 0.0.
    hgt_rate : float, optional
        Horizontal gene transfer rate, default is 0.0.
    gc_rate : float, optional
        Gene conversion rate, default is 0.0.
    dupl_polytomy : float, optional
        Allows non-binary duplication events by specifying the lambda
        parameter for a poisson distribution (copy number =
        drawn number + 2); default is 0.0.
    prohibit_extinction : str or bool, optional 
        Avoid the extinction of all members in any species ('per_species'),
        of the complete gene family ('per_family'), or no constraints
        (False); default is 'per_species'.
    replace_prob : float, optional
        Enables replacing HGT events, probability by which one random
        homolog in the receiving branch of the receiving branch gets lost
        immediately after the transfer.
    additive_transfer_distance_bias : str or bool, optional
        Specifies whether closer related species have a higher probability
        to be the recipient species in an additive HGT event. The default
        is False, in which case the recipient species is chosen at random
        among the co-existing species. The options 'inverse' and
        'exponential' mean that a species branch is sampled weighted by
        1/(a * t) or e^(-(a * t)), resp., where t is the elapsed time between
        the last common ancestor of the two species branches and the time of
        the event, see [1], and a is a user-defined factor.
    replacing_transfer_distance_bias : str or bool, optional
        Specifies whether closer related gene branches have a higher
        probability to be replaced in a replacing HGT event. The default
        is False, in which case the replaced gene is chosen at random
        among the co-existing gene branches. The options 'inverse' and
        'exponential' mean that a gene branch is sampled weighted by
        1/(a * t) or e^(-(a * t)), resp., where t is the elapsed time between
        the last common ancestor of the two gene branches and the time of the
        event, see [1], and a is a user-defined factor.
    transfer_distance_bias : str or bool, optional
        Set a common bias mode for additive and replacing HGT, see
        description of parameters 'additive_transfer_distance_bias' and
        'replacing_transfer_distance_bias'. If the latter are no set to
        the default (False), then these optioned are prioritized.
    transfer_distance_bias_strength : float, optional
        Intensity of the transfer distance bias (factor a) for additive and
        replacing HGT. The default is 1.0.
    gc_distance_bias : str or bool, optional
        Specifies whether closer related gene branches have a higher
        probability to be replaced in a gene conversion event. The default
        is False, in which case the replaced gene is chosen at random
        among the paralogs in the respective species lineage. The options
        'inverse' and 'exponential' mean that a paralog is sampled weighted
        by 1/(a * t) or e^(-(a * t)), resp., where t is the elapsed time
        between the last common ancestor of the two gene branches and the
        time of the event, see [1], and a is a user-defined factor.
    gc_distance_bias_strength : float, optional
        Intensity of the distance bias (factor a) for gene conversion.
        The default is 1.0.
    
    Returns
    -------
    Tree
        The simulated gene tree.
    
    References
    ----------
    .. [1] S. Kundu, M. S. Bansal.
       SaGePhy: an improved phylogenetic simulation framework for gene and 
       subgene evolution.
       In: Bioinformatics, 35(18), 2019, 3496–3498.
       doi:10.1093/bioinformatics/btz081.
    """
    
    gene_tree_simulator = GeneTreeSimulator(S)
    return gene_tree_simulator.simulate(**kwargs)


@dataclass
class _Branch:
    
    label:          int         # unique branch label
    list_id:        int         # index in list of branches
    parent:         TreeNode    # parent node
    S_edge:         TreeNode    # v of species tree edge (u, v) into which
                                # the branch is embedded
    transferred:    TreeNode    # 1 if HGT edge, 0 otherwise
    
    def __hash__(self):
        
        return hash(self.label)
        

class GeneTreeSimulator:
    
    def __init__(self, S):
        
        if not isinstance(S, Tree) or not S.root:
            raise TypeError("'S' must be a non-empty tree of type 'Tree'")
        
        self.S = S                                  # species tree
        self.lca_S = LCA(S)                         # lca data struct. for the species tree
        self.sorted_speciations = sorted_nodes(S)   # list of speciations sorted by time stamp
        self.sorted_edges = sorted_edges(S)         # edges of the species tree
                                                    # sort (u,v) by tstamp of u
        self._analyze_secies_tree()
    
    
    def _analyze_secies_tree(self):
        
        self.S_subtree_survivors = {}
        
        for u in self.S.postorder():
            
            if not u.children:
                if u.event == 'L':
                    self.S_subtree_survivors[u] = 0
                else:
                    self.S_subtree_survivors[u] = 1
            
            else:
                self.S_subtree_survivors[u] = sum(self.S_subtree_survivors[v]
                                                  for v in u.children)
        
        if not self.S_subtree_survivors[self.S.root]:
            warnings.warn('species tree has no non-loss leaves',
                          category=warnings.UserWarning)
    
    
    def simulate(self,
                 dupl_rate=0.0,
                 loss_rate=0.0,
                 hgt_rate=0.0,
                 gc_rate=0.0,
                 dupl_polytomy=0.0,
                 prohibit_extinction='per_species',
                 replace_prob=0.0,
                 additive_transfer_distance_bias=False,
                 replacing_transfer_distance_bias=False,
                 transfer_distance_bias=False,
                 transfer_distance_bias_strength=1.0,
                 gc_distance_bias=False,
                 gc_distance_bias_strength=1.0,
                 **kwargs):
        """Simulate a gene tree along the specified species tree.
        
        Parameters
        ----------
        dupl_rate : float, optional
            Duplication rate, default is 0.0.
        loss_rate : float, optional
            Loss rate, default is 0.0.
        hgt_rate : float, optional
            Horizontal gene transfer rate, default is 0.0.
        gc_rate : float, optional
            Gene conversion rate, default is 0.0.
        dupl_polytomy : float, optional
            Allows non-binary duplication events by specifying the lambda
            parameter for a poisson distribution (copy number =
            drawn number + 2); default is 0.0.
        prohibit_extinction : str or bool, optional 
            Avoid the extinction of all members in any species ('per_species'),
            of the complete gene family ('per_family'), or no constraints
            (False); default is 'per_species'.
        replace_prob : float, optional
            Enables replacing HGT events, probability by which one random
            homolog in the receiving branch of the receiving branch gets lost
            immediately after the transfer.
        additive_transfer_distance_bias : str or bool, optional
            Specifies whether closer related species have a higher probability
            to be the recipient species in an additive HGT event. The default
            is False, in which case the recipient species is chosen at random
            among the co-existing species. The options 'inverse' and
            'exponential' mean that a species branch is sampled weighted by
            1/(a * t) or e^(-(a * t)), resp., where t is the elapsed time
            between the last common ancestor of the two species branches and
            the time of the event, see [1], and a is a user-defined factor.
        replacing_transfer_distance_bias : str or bool, optional
            Specifies whether closer related gene branches have a higher
            probability to be replaced in a replacing HGT event. The default
            is False, in which case the replaced gene is chosen at random
            among the co-existing gene branches. The options 'inverse' and
            'exponential' mean that a gene branch is sampled weighted by
            1/(a * t) or e^(-(a * t)), resp., where t is the elapsed time
            between the last common ancestor of the two gene branches and the
            time of the event, see [1], and a is a user-defined factor.
        transfer_distance_bias : str or bool, optional
            Set a common bias mode for additive and replacing HGT, see
            description of parameters 'additive_transfer_distance_bias' and
            'replacing_transfer_distance_bias'. If the latter are no set to
            the default (False), then these optioned are prioritized.
        transfer_distance_bias_strength : float, optional
            Intensity of the transfer distance bias (factor a) for additive and
            replacing HGT. The default is 1.0.
        gc_distance_bias : str or bool, optional
            Specifies whether closer related gene branches have a higher
            probability to be replaced in a gene conversion event. The default
            is False, in which case the replaced gene is chosen at random
            among the paralogs in the respective species lineage. The options
            'inverse' and 'exponential' mean that a paralog is sampled weighted
            by 1/(a * t) or e^(-(a * t)), resp., where t is the elapsed time
            between the last common ancestor of the two gene branches and the
            time of the event, see [1], and a is a user-defined factor.
        gc_distance_bias_strength : float, optional
            Intensity of the distance bias (factor a) for gene conversion.
            The default is 1.0.
        
        Returns
        -------
        Tree
            The simulated gene tree.
        
        References
        ----------
        .. [1] S. Kundu, M. S. Bansal.
           SaGePhy: an improved phylogenetic simulation framework for gene and 
           subgene evolution.
           In: Bioinformatics, 35(18), 2019, 3496–3498.
           doi:10.1093/bioinformatics/btz081.
        """
        
        self.rate_sum = 0.0
        for rate in (dupl_rate, loss_rate, hgt_rate, gc_rate):
            self.rate_sum += rate
            if rate < 0.0:
                raise ValueError(f'negative rate {rate}')
        
        self.d = dupl_rate
        self.l = loss_rate
        self.h = hgt_rate
        self.gc = gc_rate
        
        if prohibit_extinction not in (False, 'per_family', 'per_species'):
            raise ValueError(f"unknown mode prohibit_extinction attribute: "\
                             f"{prohibit_extinction}")
        self._prohibit_extinction = prohibit_extinction
        
        self._dupl_polytomy = dupl_polytomy
        
        self._replace_prob = replace_prob
        
        for m in (additive_transfer_distance_bias,
                  replacing_transfer_distance_bias,
                  transfer_distance_bias,
                  gc_distance_bias):
            if m not in (False, 'inverse', 'exponential'):
                raise ValueError(f"unknown mode for transfer distance bias: "\
                                 f"{m}")
        
        # mode for transfer distance bias for additive and replacing HGT
        self._additive_transfer_distance_bias = additive_transfer_distance_bias
        self._replacing_transfer_distance_bias = replacing_transfer_distance_bias
        if transfer_distance_bias:
            if not additive_transfer_distance_bias:
                self._additive_transfer_distance_bias = transfer_distance_bias
            if not replacing_transfer_distance_bias:
                self._replacing_transfer_distance_bias = transfer_distance_bias
        
        # intensity factor for transfer distance bias
        if not isinstance(transfer_distance_bias_strength, (int, float)) or \
            transfer_distance_bias_strength <= 0.0:
            raise ValueError('factor for transfer distance bias must be > 0')
        self._transfer_distance_bias_strength = transfer_distance_bias_strength
        
        self._gc_distance_bias = gc_distance_bias
        if not isinstance(gc_distance_bias_strength, (int, float)) or \
            gc_distance_bias_strength <= 0.0:
            raise ValueError('factor for g.c. distance bias must be > 0')
        self._gc_distance_bias_strength = gc_distance_bias_strength
        
        return self._run()
    
    
    def _new_branch(self, parent, S_edge, transferred):
        
        new_branch = _Branch(self.id_counter, len(self.branches),
                             parent, S_edge, transferred)
        self.branches.append(new_branch)
        
        self.ES_to_b[S_edge].append(new_branch)
        self.id_counter += 1
        
        if self.S_subtree_survivors[S_edge]:
            self.surv_non_loss_lineages.add(new_branch)
        
        return new_branch
    
    
    def _remove_branch(self, branch):
        
        # efficient removal from list
        if branch.list_id != len(self.branches) - 1:
            branch2 = self.branches[-1]
            self.branches[branch.list_id] = branch2
            branch2.list_id = branch.list_id
        self.branches.pop()
        
        self.ES_to_b[branch.S_edge].remove(branch)
        self.surv_non_loss_lineages.discard(branch)
    
    
    def _initiatialize_tree(self):
        
        root = TreeNode(label=0, 
                        event='S' if len(self.S.root.children) > 1 else None,
                        reconc=self.S.root.label, 
                        tstamp=self.S.root.tstamp)
            
        T = Tree(root)
        self.id_counter += 1
        self.spec_queue.popleft()
                
        for S_edge in self.S.root.children:
            self._new_branch(T.root, S_edge, 0)
            
        return T
    
    
    def _get_branch_and_type(self):
        
        total_rate = len(self.branches) * self.rate_sum
        r = np.random.uniform(high=total_rate)
        i = int(len(self.branches) * r / total_rate)
        r2 = r - i * self.rate_sum
        
        if r2 <= self.d:
            event_type = 'D'
        elif r2 <= self.d + self.l:
            event_type = 'L'
        elif r2 <= self.d + self.l + self.h:
            event_type = 'H'
        else:
            event_type = 'GC'
        
        return self.branches[i], event_type


    def _run(self):
        
        # queue for speciation events
        self.spec_queue = deque(self.sorted_speciations)
        self.id_counter = 0
        
        # keep track of surving branches that are in species branches with
        # at least 1 surviving species leaf
        self.surv_non_loss_lineages = set()
        
        self.branches = []
        
        # maps species tree branches to existing gene branches
        self.ES_to_b = {S_edge: [] for _, S_edge in self.sorted_edges}
        
        self.T = self._initiatialize_tree()
        t = self.T.root.tstamp
        
        while self.spec_queue:
            
            total_rate = len(self.branches) * self.rate_sum
            event_tstamp = t - np.random.exponential(1/total_rate) \
                           if total_rate > 0.0 else -1
            next_spec_tstamp = self.spec_queue[0].tstamp
            
            # speciation
            if event_tstamp <= next_spec_tstamp:
                self._speciation()      
                t = next_spec_tstamp
                
            else:
                branch, event_type = self._get_branch_and_type()
                
                # duplication
                if event_type == 'D':
                    self._duplication(event_tstamp, branch)
                
                # loss
                elif event_type == 'L':
                    self._loss(event_tstamp, branch)
                    
                # HGT    
                elif event_type == 'H':
                    self._hgt(event_tstamp, branch)
                
                # gene conversion
                elif event_type == 'GC':
                    self._gene_conversion(event_tstamp, branch)
                
                t = event_tstamp
        
        # add the 'dist' attribute to the nodes
        distance_from_timing(self.T)

        return self.T
            
    
    def _speciation(self):
        
        # also handles loss and leaf nodes of the species tree
        
        S_edge = self.spec_queue.popleft()
        
        # copy since we modify this list
        for branch in self.ES_to_b[S_edge].copy():
            
            spec_node = TreeNode(label=branch.label,
                                 event='S',
                                 reconc=S_edge.label, tstamp=S_edge.tstamp,
                                 transferred=branch.transferred)
            branch.parent.add_child(spec_node)
            self._remove_branch(branch)
            
            for S_w in S_edge.children:
                self._new_branch(spec_node, S_w, 0)
            
            # loss leaves if it was a species extinction event
            if (not S_edge.children) and S_edge.label == 'L':
                spec_node.event = 'L'
            
    
    def _duplication(self, event_tstamp, branch):
        
        S_edge = branch.S_edge
        
        dupl_node = TreeNode(label=branch.label,
                             event='D',
                             reconc=(S_edge.parent.label, S_edge.label),
                             tstamp=event_tstamp,
                             transferred=branch.transferred)
        branch.parent.add_child(dupl_node)
        self._remove_branch(branch)
        
        copy_number = 2 if self._dupl_polytomy <= 0.0 else \
                      2 + np.random.poisson(lam=self._dupl_polytomy)
        
        for i in range(copy_number):
            self._new_branch(dupl_node, S_edge, 0)
            
            
    def _loss(self, event_tstamp, branch):
        
        # not executing the loss event if extinction is prohibited is
        # equivalent to temporarily setting the loss rate to zero in the
        # respective branches
        if (self._prohibit_extinction == 'per_family' and
            len(self.surv_non_loss_lineages) == 1 and
            next(iter(self.surv_non_loss_lineages)) is branch):
            return
        
        if (self._prohibit_extinction == 'per_species' and
            len(self.ES_to_b[branch.S_edge]) <= 1):
            return
        
        S_edge = branch.S_edge
        
        loss_node = TreeNode(label=branch.label,
                             event='L',
                             reconc=(S_edge.parent.label, S_edge.label),
                             tstamp=event_tstamp,
                             transferred=branch.transferred)
        branch.parent.add_child(loss_node)
        self._remove_branch(branch)
            
    
    def _coexisting_species_edges(self, tstamp, exclude_edge=None):
        """Return list of edges for the given timestamp."""
        
        valid_species = []
        
        for S_u, S_v in self.sorted_edges:
            
            if S_u.tstamp <= tstamp:
                break
            elif S_v.tstamp < tstamp and S_v != exclude_edge:
                valid_species.append(S_v)
                
        return valid_species
    
    
    def _sample_recipient(self, event_tstamp, branch):
        
        S_edge = branch.S_edge
        trans_edge, replaced_gene_branch = None, None
        
        valid_species = self._coexisting_species_edges(event_tstamp,
                                                       exclude_edge=S_edge)
        
        if not valid_species:
            return None, None
        
        # ---- ADDITIVE HGT EVENT ----
        if np.random.random() >= self._replace_prob:
            
            # ---- no transfer distance bias ---
            if not self._additive_transfer_distance_bias:
                trans_edge = random.choice(valid_species)
            # ---- transfer distance bias ---
            else:
                distances = [(self.lca_S(S_edge, e).tstamp-event_tstamp)
                             for e in valid_species]
                a = self._transfer_distance_bias_strength
                if self._additive_transfer_distance_bias == 'inverse':
                    weights = 1 / (a * np.asarray(distances))
                elif self._additive_transfer_distance_bias == 'exponential':
                    weights = np.exp(-a * np.asarray(distances))
                
                trans_edge = random.choices(valid_species, weights=weights)[0]
            
        # ---- REPLACING HGT EVENT ----
        else:
            valid_genes = [b for e in valid_species for b in self.ES_to_b[e]]
            
            if not valid_genes:
                return None, None
            
            # ---- no transfer distance bias ---
            if not self._replacing_transfer_distance_bias:
                replaced_gene_branch = random.choice(valid_genes)
            # ---- transfer distance bias ---
            else:
                lca_T = LCA(self.T)
                distances = [(lca_T(branch.parent, b.parent).tstamp-event_tstamp)
                             for b in valid_genes]
                a = self._transfer_distance_bias_strength
                if self._replacing_transfer_distance_bias == 'inverse':
                    weights = 1 / (a * np.asarray(distances))
                elif self._replacing_transfer_distance_bias == 'exponential':
                    weights = np.exp(-a * np.asarray(distances))
                
                replaced_gene_branch = random.choices(valid_genes,
                                                      weights=weights)[0]
            
            trans_edge = replaced_gene_branch.S_edge
        
        return trans_edge, replaced_gene_branch
    
    
    def _hgt(self, event_tstamp, branch):
        
        S_edge = branch.S_edge
        
        trans_edge, replaced_gene_branch = self._sample_recipient(event_tstamp,
                                                                  branch)
        
        if trans_edge:
            hgt_node = TreeNode(label=branch.label,
                                event='H',
                                reconc=(S_edge.parent.label, S_edge.label),
                                tstamp=event_tstamp,
                                transferred=branch.transferred)
            branch.parent.add_child(hgt_node)
            self._remove_branch(branch)
            
            # original branch
            self._new_branch(hgt_node, S_edge, 0)
            
            # receiving branch
            self._new_branch(hgt_node, trans_edge, 1)
        
        # replacing HGT leads to loss in the recipient species
        if replaced_gene_branch:
            self._loss(event_tstamp, replaced_gene_branch)
            # save replaced gene information in HGT node
            hgt_node.replaced_gene = replaced_gene_branch.label
    
    
    def _sample_replaced_gene_gc(self, event_tstamp, branch):
        
        candidates = [b for b in self.ES_to_b[branch.S_edge] if b is not branch]
        
        # no gene conversion possible if there is currently no other branch in
        # the species lineage
        if not candidates:
            return
        
        if not self._gc_distance_bias:
            return random.choice(candidates)
        else:
            lca_T = LCA(self.T)
            distances = [(lca_T(branch.parent, b.parent).tstamp-event_tstamp)
                         for b in candidates]
            a = self._gc_distance_bias_strength
            if self._gc_distance_bias == 'inverse':
                weights = 1 / (a * np.asarray(distances))
            elif self._gc_distance_bias == 'exponential':
                weights = np.exp(-a * np.asarray(distances))
            
            return random.choices(candidates, weights=weights)[0]
    
    
    def _gene_conversion(self, event_tstamp, branch):
        
        S_edge = branch.S_edge
        
        replaced_gene_branch = self._sample_replaced_gene_gc(event_tstamp,
                                                             branch)
        
        if replaced_gene_branch:
            gc_node = TreeNode(label=branch.label,
                               event='GC',
                               reconc=(S_edge.parent.label, S_edge.label),
                               tstamp=event_tstamp,
                               transferred=branch.transferred)
            branch.parent.add_child(gc_node)
            self._remove_branch(branch)
            
            self._new_branch(gc_node, S_edge, 0)
            self._new_branch(gc_node, S_edge, 0)
        
            # gene conversion leads to loss of a homologous gene
            self._loss(event_tstamp, replaced_gene_branch)
            # save replaced gene information in GC node
            gc_node.replaced_gene = replaced_gene_branch.label
        

# --------------------------------------------------------------------------
#                           PRUNE LOSS BRANCHES
# --------------------------------------------------------------------------

def prune_losses(tree):
    """Prune all loss branches.
    
    Returns a copy of the tree with all branches that lead to losses only
    removed and superfluous vertices suppressed. Additionally, if the root of
    the tree has only a single child, then this 'planted edge' is also removed.    
    
    Parameters
    ----------
    tree : Tree
        The tree to be pruned.
    
    Returns
    -------
    Tree
        A pruned version of the tree.
    """
    
    pruned_tree = delete_losses_and_contract(tree, inplace=False)
    
    remove_planted_root(pruned_tree, inplace=True)
    
    return pruned_tree


def observable_tree(tree):
    
    warn('This method is deprecated. Use prune_losses() instead.',
         DeprecationWarning, stacklevel=2)
    return prune_losses(tree)



# -*- coding: utf-8 -*-

"""
Orthology graph, BMG, RBMG.

This module provides classes concerning colored graphs and colored
(phylogenetic) trees, including a BMG(Tree)-function, and computation of an
LRT from a (not necessarily valid) BMG or from a (pruned) gene tree.

Phylogenetic best matches of a gene x of species X are defined as those genes
y of another species Y that share the lowest common ancestor with x in the gene
tree among all genes in that species. In contrast, two genes are orthologs if
their last common ancestor was a speciation event. Orthology and reciprocal
best matches are closely related. The module asymmetree.analysis. BestMatches
contains methods for extracting BMGs from trees, least resolved trees of BMG,
and orthology graphs from trees.
"""

import itertools

import networkx as nx

from tralda.datastructures.Tree import Tree, TreeNode

from tralda.tools.GraphTools import (sort_by_colors,
                                     is_properly_colored,
                                     graphs_equal)
from tralda.supertree import Build

from asymmetree.tools.PhyloTreeTools import (topology_only,
                                             reconstruct_reconc_from_graph)


__author__ = 'David Schaller'


# --------------------------------------------------------------------------
#                 True Orthology Graph (TOG), BMG & RBMG
#
#                         (given a known tree)
# --------------------------------------------------------------------------

def orthology_from_tree(tree):
    """Constructs the true orthology graph from an event-labeled tree.
    
    Parameters
    ----------
    tree : tree
        A tree with leaves that have the `label`, and inner vertices that have
        the `event` attribute set.
    
    Returns
    -------
    networx.Graph
        The ortholy graph explained by the tree, i.e., a graph on the set of
        leaf labels and edges between two of them if and only if their last
        common ancestor was a speciation event.
    """
    
    if not isinstance(tree, Tree):
        raise TypeError("not of type 'Tree'")
    
    leaves = tree.leaf_dict()
    TOG = nx.Graph()
    
    for v in leaves[tree.root]:
        TOG.add_node(v.label, color=v.reconc)
    
    for node in tree.preorder():
        if node.event == 'S':
            for c1, c2 in itertools.combinations(node.children, 2):
                for u in leaves[c1]:
                    for v in leaves[c2]:
                        TOG.add_edge(u.label, v.label)
                        
    return TOG


def bmg_from_tree(tree, supply_rbmg=False):
    """Construct a BMG (and optionally RBMG) from a given tree.
    
    Parameters
    ----------
    tree : tree
        A tree with leaves that have the `label` and `reconc` attribute set.
    supply_rbmg : bool, optional
        If True, also return the symmetric part of the constructed BMG.
        The default is False.
    
    Returns
    -------
    networx.DiGraph or pair of networx.DiGraph and networx.Graph
        The constructed BMG and optionally its symmetric part.
    
    References
    ----------
    .. [1] M. Geiß, E. Chávez, M. González Laffitte, A. López Sánchez, B. M. R. 
       Stadler, D. I. Valdivia, M. Hellmuth, M. Hernández Rosales, and P. F.
       Stadler.
       Best match graphs.
       In: Journal of Mathematical Biology, 78:2015–2057, 2019.
       doi:10.1007/s00285-019-01332-9.
    """
    
    if not isinstance(tree, Tree):
        raise TypeError("not of type 'Tree'")
    
    leaves = tree.leaf_dict()
    bmg = nx.DiGraph()
    colors = set()
    
    for v in leaves[tree.root]:
        colors.add(v.reconc)
        bmg.add_node(v.label, color=v.reconc)
    
    for u in leaves[tree.root]:
        remaining = colors - set([u.reconc])            # colors to which no best match has yet been found
        parent = u.parent                               # start with direct parent of each node
        while remaining and parent:
            colors_here = set()
            for v in leaves[parent]:
                if v.reconc in remaining:               # best match found
                    colors_here.add(v.reconc)
                    bmg.add_edge(u.label, v.label)      # insert edge (u,v)
            remaining -= colors_here                    # update remaining colors
            parent = parent.parent
    
    if not supply_rbmg:
        return bmg
    else:
        return bmg, bmg.to_undirected(reciprocal=True)
    
    
def extended_best_hits(leaves, D, epsilon=1e-8, supply_rbmg=False):
    """Compute BMG and RBMG from a distances matrix D.
    
    Apporximation of best matches as the genes within a tolerance range from
    the gene with the smallest distance, see [1].
    
    Parameters
    ----------
    leaves : list of TreeNode instances with `label` attribute
    D : 2-dimensional numpy array
        The distances of the leaves where the order of the rows/columns
        corresponds to that in the list.
    epsilon : float, optional
        relative threshold for the inclusion of suboptimal hits, i.e.,
        (x,y) in BMG if D(x,y) <= (1+epsilon) * min d(x,y'). The default
        is 1e-8 (for limited float precision).
    supply_rbmg : bool, optional
        If True, also return the symmetric part of the constructed BMG.
        The default is False.
    
    Returns
    -------
    networx.DiGraph or pair of networx.DiGraph and networx.Graph
        The constructed BMG and optionally its symmetric part.
        
    References
    ----------
    .. [1] P. F. Stadler, M. Geiß, D. Schaller, A. López Sánchez, M. Gonzalez
    Laffitte, D. I. Valdivia, M. Hellmuth, and M. Hernández Rosales.
    From pairs of most similar sequences to phylogenetic best matches.
    In: Algorithms for Molecular Biology, 15:5, 2020.
    doi: 10.1186/s13015-020-00165-2.
    """
    
    bmg = nx.DiGraph()
    colors = set()
    relative_threshold = 1 + epsilon
    
    for v in leaves:
        bmg.add_node(v.label, color=v.reconc)
        colors.add(v.reconc)
    
    # ---- build BMG ----
    for i, u in enumerate(leaves):
        minima = {c: float('inf') for c in colors}
        for j, v in enumerate(leaves):
            if D[i,j] < minima[v.reconc]:
                minima[v.reconc] = D[i,j]
        for j, v in enumerate(leaves):
            if (u.reconc != v.reconc and
                D[i,j] <= relative_threshold * minima[v.reconc]):
                bmg.add_edge(u.label, v.label, distance = D[i,j])
    
    if not supply_rbmg:
        return bmg
    else:
        return bmg, bmg.to_undirected(reciprocal=True)


# --------------------------------------------------------------------------
#                         Least resolved trees
# --------------------------------------------------------------------------

def informative_triples(graph, color_dict=None):
    """Compute the informative triples of a colored digraph.
    
    Parameters
    ----------
    graph : networkx.DiGraph
        A digraph whose nodes have the 'color' attribute.
    color_dict : dict, optional
        A dict containing the colors that appear in the graph as keys and the
        list of nodes with the respective color as values. The default is False
        in which case this dict is computed internally.
    
    Returns
    -------
    list
        Each item in the list is a tuple (a, b1, b2) of three nodes such that
        a b1 | b2 is an informative triple in the graph, see [1].
    
    References
    ----------
    .. [1] M. Geiß, E. Chávez, M. González Laffitte, A. López Sánchez, B. M. R. 
       Stadler, D. I. Valdivia, M. Hellmuth, M. Hernández Rosales, and P. F.
       Stadler.
       Best match graphs.
       In: Journal of Mathematical Biology, 78:2015–2057, 2019.
       doi:10.1007/s00285-019-01332-9.
    """
    
    if not isinstance(graph, nx.DiGraph):
        raise TypeError("must be a NetworkX 'Digraph'")
    
    if not color_dict:
        color_dict = sort_by_colors(graph)
    
    R = []
    
    for c1, c2 in itertools.permutations(color_dict.keys(), 2):
        for a in color_dict[c1]:
            for b1, b2 in itertools.permutations(color_dict[c2], 2):
                if graph.has_edge(a, b1) and (not graph.has_edge(a, b2)):
                    R.append( (a, b1, b2) )
    
    return R


def forbidden_triples(graph, color_dict=None):
    """Compute the forbidden triples of a colored digraph.
    
    Parameters
    ----------
    graph : networkx.DiGraph
        A digraph whose nodes have the 'color' attribute.
    color_dict : dict, optional
        A dict containing the colors that appear in the graph as keys and the
        list of nodes with the respective color as values. The default is False
        in which case this dict is computed internally.
    
    Returns
    -------
    list
        Each item in the list is a tuple (a, b1, b2) of three nodes such that
        a b1 | b2 is a forbidden triple in the graph, see [1].
    
    References
    ----------
    .. [1] D. Schaller, P. F. Stadler, and M. Hellmuth.
       Complexity of modification problems for best match graphs.
       In: Theoretical Computer Science, 865:63–84, 2021.
       doi:10.1016/j.tcs.2021.02.037.
    """
    
    if not isinstance(graph, nx.DiGraph):
        raise TypeError("must be a NetworkX 'Digraph'")
    
    if not color_dict:
        color_dict = sort_by_colors(graph)
    
    F = []
    
    for c1, c2 in itertools.permutations(color_dict.keys(), 2):
        for a in color_dict[c1]:
            for b1, b2 in itertools.combinations(color_dict[c2], 2):
                if graph.has_edge(a, b1) and graph.has_edge(a, b2):
                    F.append( (a, b1, b2) )
                    F.append( (a, b2, b1) )
    
    return F


def informative_forbidden_triples(graph, color_dict=None):
    """Compute the informative and forbidden triples of a colored digraph.
    
    Parameters
    ----------
    graph : networkx.DiGraph
        A digraph whose nodes have the 'color' attribute.
    color_dict : dict, optional
        A dict containing the colors that appear in the graph as keys and the
        list of nodes with the respective color as values. The default is False
        in which case this dict is computed internally.
    
    Returns
    -------
    tuple of two lists
        Each item in the first (resp. second) list is a tuple (a, b1, b2) of
        three nodes such that a b1 | b2 is a informative (resp. forbidden)
        triple in the graph, see [1].
    
    References
    ----------
    .. [1] D. Schaller, P. F. Stadler, and M. Hellmuth.
       Complexity of modification problems for best match graphs.
       In: Theoretical Computer Science, 865:63–84, 2021.
       doi:10.1016/j.tcs.2021.02.037.
    """
    
    if not isinstance(graph, nx.DiGraph):
        raise TypeError("must be a NetworkX 'Digraph'")
    
    if not color_dict:
        color_dict = sort_by_colors(graph)
    
    R, F = [], []
    
    for c1, c2 in itertools.permutations(color_dict.keys(), 2):
        for a in color_dict[c1]:
            for b1, b2 in itertools.combinations(color_dict[c2], 2):
                
                if graph.has_edge(a, b1) and graph.has_edge(a, b2):
                    F.append( (a, b1, b2) )
                    F.append( (a, b2, b1) )
                elif graph.has_edge(a, b1):
                    R.append( (a, b1, b2) )
                elif graph.has_edge(a, b2):
                    R.append( (a, b2, b1) )
    
    return R, F


def binary_explainable_triples(graph, color_dict=None):
    """Extended informative triple set for binary-explainable graphs.
    
    Parameters
    ----------
    graph : networkx.DiGraph
        A digraph whose nodes have the 'color' attribute.
    color_dict : dict, optional
        A dict containing the colors that appear in the graph as keys and the
        list of nodes with the respective color as values. The default is False
        in which case this dict is computed internally.
    
    Returns
    -------
    tuple of two lists
        Each item in the  list is a tuple (a, b, c) of three nodes such that
        ab|c is a triple displayed by every binary tree that explains the graph
        (if existent), see [1].
    
    References
    ----------
    .. [1] D. Schaller, M. Geiß, M. Hellmuth, and P. F. Stadler.
       Best match graphs with binary trees.
       In: C. Martín-Vide, M. A. Vega-Rodríguez, and T. Wheeler, editors,
       Algorithms for Computational Biology, 8th AlCoB, volume 12715 of Lecture
       Notes in Computer Science, pages 82–93, 2021.
       doi: 10.1007/978-3-030-74432-8_6.
    """
    
    if not isinstance(graph, nx.DiGraph):
        raise TypeError("must be a NetworkX 'Digraph'")
    
    if not color_dict:
        color_dict = sort_by_colors(graph)
    
    R_binary = []
    
    for c1, c2 in itertools.permutations(color_dict.keys(), 2):
        for a in color_dict[c1]:
            for b1, b2 in itertools.combinations(color_dict[c2], 2):
                
                if graph.has_edge(a, b1) and graph.has_edge(a, b2):
                    try:
                        if b2 < b1:
                            b1, b2 = b2, b1
                    except NotImplemented: pass
                    R_binary.append( (b1, b2, a) )
                elif graph.has_edge(a, b1):
                    R_binary.append( (a, b1, b2) )
                elif graph.has_edge(a, b2):
                    R_binary.append( (a, b2, b1) )
    
    return R_binary


def _finalize(tree, G):
    
    if not tree:
        return None
    
    if isinstance(tree, TreeNode):
        tree = Tree(tree)
    
    # assign reconciliationss to the leaves
    reconstruct_reconc_from_graph(tree, G)
    
    return tree


# --------------------------------------------------------------------------
#                      LRT FROM OBSERVABLE GENE TREE
# --------------------------------------------------------------------------

def lrt_from_tree(T):
    """Computes the Least Resolved Tree from a tree.
    
    The unique Least Resolved Tree from a (pruned) gene tree is computed by
    contraction of all redundant edges.
    
    Parameters
    ----------
    T : Tree
        A tree whose leaf nodes have the 'reconc' attribute.
    
    Returns
    -------
    Tree
        The unique least resolved tree (w.r.t. best matches), see [1].
    
    References
    ----------
    .. [1] D. Schaller, M. Geiß, P. F. Stadler, and M. Hellmuth.
    Complete characterization of incorrect orthology assignments in best match
    graphs. 
    In: Journal of Mathematical Biology, 82:20, 2021.
    doi: 10.1007/s00285-021-01564-8.
    """
    
    lrt = T.copy()
    if not lrt.root:
        return lrt
    
    # remove planted root if existent
    if len(lrt.root.children) == 1:
        new_root = lrt.root.children[0]
        new_root.detach()
        lrt.root = new_root
        new_root.dist = 0.0
    
    # assign list of leaves to each node
    leaves = lrt.leaf_dict()
    
    subtree_reconcs = {}
    for v in lrt.preorder():
        subtree_reconcs[v] = {leaf.reconc for leaf in leaves[v]}
        
    arc_colors = _arc_colors(lrt, leaves, subtree_reconcs)
    red_edges = redundant_edges(lrt, subtree_reconcs, arc_colors)
    lrt.contract(red_edges)
    lrt = topology_only(lrt)
    
    return lrt


def _arc_colors(T, leaves, subtree_reconcs):
    """Color sets relevant for redundant edge computation.
    
    Computes for all inner vertices v the color set of y such that y with (x,y)
    is an arc in the BMG and lca(x,y) = v.
    """
    
    all_colors = subtree_reconcs[T.root]
    
    # color sets for all v
    arc_colors = {v: set() for v in T.preorder()}
    
    for u in leaves[T.root]:
        
        # colors to which no best match has yet been found
        remaining = all_colors - {u.reconc}
        
        # start with direct parent of each node
        current = u.parent
        
        while remaining and current:
            colors_here = set()
            for v in leaves[current]:
                
                # best match found
                if v.reconc in remaining:
                    colors_here.add(v.reconc)
            
            arc_colors[current].update(colors_here)
            remaining -= colors_here
            current = current.parent
    
    return arc_colors


def redundant_edges(T, subtree_reconcs, arc_colors):
    
    red_edges = []
    
    for u, v in T.inner_edges():
        
        # colors s in sigma( L(T(u) \ T(v)) )
        aux_set = set()
        
        for v2 in u.children:
            if v2 is not v:
                aux_set.update(subtree_reconcs[v2])
        
        if not arc_colors[v].intersection(aux_set):
            red_edges.append((u, v))
    
    return red_edges


# --------------------------------------------------------------------------
#                               LRT FROM BMG
# --------------------------------------------------------------------------

def is_bmg(G):
    """Determine whether a colored digraph is a BMG.
    
    If the graph is a BMG, then its LRT is returned, see [1].
    
    Parameters
    ----------
    G : networkx.DiGraph
        A digraph whose nodes have the 'color' attribute.
    
    Returns
    -------
    Tree or bool
        The least-resolved tree if the graph was a BMG, and False otherwise.
    
    References
    ----------
    .. [1] D. Schaller, M. Geiß, E. Chávez, M. González Laffitte, A. López
       Sánchez, B. M. R. Stadler, D. I. Valdivia, M. Hellmuth, M. Hernández
       Rosales, and P. F. Stadler.
       Corrigendum to 'Best match graphs'.
       In: Journal of Mathematical Biology, 82:47, 2021. 
       doi: 10.1007/s00285-021-01601-6.
    """
    
    if not isinstance(G, nx.DiGraph):
            raise TypeError('not a digraph')
    if not is_properly_colored(G):
        raise RuntimeError('not a properly colored digraph')
    
    subtrees = []
    colors = None
    
    for wcc in nx.weakly_connected_components(G):
        
        sg = G.subgraph(wcc).copy()
        color_dict = sort_by_colors(sg)
        
        # the color sets of all components must be equal
        if colors is None:
            colors = set(color_dict)
        elif colors != set(color_dict):
            return False
        
        L = {v for v in sg.nodes()}
        R = informative_triples(sg, color_dict=color_dict)
        build = Build(R, L, mincut=False)
        subtree = build.build_tree()
        
        if not subtree: 
            return False
        else:
            # a digraph is a BMG iff its equal to the BMG of BUILD(R)
            reconstruct_reconc_from_graph(subtree, sg)
            if not graphs_equal(sg, bmg_from_tree(subtree)):
                return False
            subtrees.append(subtree)
    
    if len(subtrees) == 1:
        root = subtrees[0].root
    else:
        root = TreeNode()
        for subtree in subtrees:
            root.add_child(subtree.root)
    
    return _finalize(root, G)
    

def lrt_from_colored_graph(G, mincut=False, weighted_mincut=False):
    
    L = {v for v in G.nodes()}
    R = informative_triples(G)
    
    build = Build(R, L, mincut=mincut, weighted_mincut=weighted_mincut)
    tree = build.build_tree()
    
    return _finalize(tree, G)


def correct_bmg(bmg_original):
    """Build the LRT (using min cut in BUILD algorithm) and return its BMG."""
    
    subtrees = []
    for sg in (bmg_original.subgraph(c)
               for c in nx.weakly_connected_components(bmg_original)):
        tree = lrt_from_colored_graph(sg, mincut=True)
        if tree:
            subtrees.append(tree)
            
    if len(subtrees) == 0:
        return None
    elif len(subtrees) == 1:
        tree = subtrees[0]
    else:
        tree = Tree(TreeNode())
        for subtree in subtrees:
            tree.root.add_child(subtree.root)
    
    return bmg_from_tree(tree)


# --------------------------------------------------------------------------
#                          LRT FROM 2-col. BMG
#                         (new characterization)
# --------------------------------------------------------------------------
            
class TwoColoredLRT:
    """Construct the least-resolved tree of a 2-colored best match graph.
    
    Implementation of the efficient algorithm in [1].
    
    References
    ----------
    .. [1] D. Schaller, M. Geiß, M. Hellmuth, and P. F. Stadler.
       Least resolved trees for two-colored best match graphs.
       In: Journal of Graph Algorithms and Applications, 25(1):397–416, 2021.
       doi: 10.7155/jgaa.00564.
    """
    
    def __init__(self, digraph):
        """
        Parameters
        ----------
        G : networkx.DiGraph
            A digraph whose nodes have the 'color' attribute.
        """
        
        if not isinstance(digraph, nx.DiGraph):
            raise TypeError('not a digraph')
            
        self.digraph = digraph
        self.color_dict = sort_by_colors(digraph)
                
    
    def build_tree(self):
        """Construct the least-resolved tree.
        
        Returns
        -------
        Tree or bool
            The least-resolved tree if the graph was a 2-colored BMG, and
            False otherwise.
        """
        
        if not is_properly_colored(self.digraph):
            raise RuntimeError('not a properly colored digraph')
        if len(self.color_dict) > 2:
            raise RuntimeError('more than 2 colors')
        
        # star tree if there is only one color
        if len(self.color_dict) == 1:
            root = TreeNode()
            for v in self.digraph.nodes():
                root.add_child(TreeNode(label=v))
        # 2 colors
        else:
            subtrees = []
            for wcc in nx.weakly_connected_components(self.digraph):
                if len(wcc) == 1:
                    return False
                
                subroot = self._build_tree(self.digraph.subgraph(wcc).copy())
                
                if not subroot:
                    return False
                else:
                    subtrees.append(subroot)
            
            if len(subtrees) == 1:
                root = subtrees[0]
            else:
                root = TreeNode()
                for subroot in subtrees:
                    root.add_child(subroot)
        
        return _finalize(root, self.digraph)
    
        
    def _build_tree(self, G):
            
        color_count = {c: 0 for c in self.color_dict.keys()}
        for v in G.nodes():
            color_count[G.nodes[v]['color']] += 1
        other_color = {color: G.order() - count
                       for color, count in color_count.items()}
        
        umbrella = {v for v in G.nodes() if 
                    other_color[G.nodes[v]['color']] == G.out_degree(v)}
        S_1 = {v for v in umbrella if umbrella.issuperset(G.predecessors(v))}
        S_2 = {v for v in S_1 if S_1.issuperset(G.predecessors(v))}
        
        if not S_2 or len(S_1) != len(S_2):
            return False
            
        node = TreeNode()
        for v in S_2:
            node.add_child(TreeNode(label=v))
            G.remove_node(v)
        
        for wcc in nx.weakly_connected_components(G):
            
            if len(wcc) == 1:
                return False
            
            child = self._build_tree(G.subgraph(wcc).copy())
            
            if not child:
                return False
            else:
                node.add_child(child)
                
        return node
    

def lrt_from_2bmg(G):
    """Effieciently constructs the LRT for a 2-BMG via the support vertices.
    
    Implementation of the efficient algorithm in [1].
    
    Parameters
    ----------
    G : networkx.DiGraph
        A digraph whose nodes have the 'color' attribute.
    
    Returns
    -------
    Tree or bool
        The least-resolved tree if the graph was a 2-colored BMG, and
        False otherwise.
    
    References
    ----------
    .. [1] D. Schaller, M. Geiß, M. Hellmuth, and P. F. Stadler.
       Least resolved trees for two-colored best match graphs.
       In: Journal of Graph Algorithms and Applications, 25(1):397–416, 2021.
       doi: 10.7155/jgaa.00564.
    """
    
    builder = TwoColoredLRT(G)
    return builder.build_tree()


# --------------------------------------------------------------------------
#                      Binary-refinable tree (BRT)
# --------------------------------------------------------------------------

def binary_refinable_tree(G, mincut=False, weighted_mincut=False):
    """Compute the binary-refinable tree (BRT) for a graph.
    
    The BRT of a BMG can, if it exists, be refined arbitrarily and such that
    it still explains the BMG. In particular, all binary explainations are 
    refinements of the BRT.
    
    Parameters
    ----------
    G : networkx.DiGraph
        A digraph whose nodes have the 'color' attribute.
    mincut : bool, optional
        Handle inconsistencies that occur if the supplied graph is not a
        binary-explainable BMG. The default is True.
    weighted_mincut : bool, optional
        If True, apply a weighted mincut heuristic. The deafault is False.
    
    Returns
    -------
    Tree or bool
        The unique binary-resovable tree if the graph is a binary-explainable
        BMG, see [1], or a heuristic tree if it is not but 'mincut' is True,
        or False if neither is the case.
    
    References
    ----------
    .. [1] D. Schaller, M. Geiß, M. Hellmuth, and P. F. Stadler.
       Best match graphs with binary trees.
       In: C. Martín-Vide, M. A. Vega-Rodríguez, and T. Wheeler, editors,
       Algorithms for Computational Biology, 8th AlCoB, volume 12715 of Lecture
       Notes in Computer Science, pages 82–93, 2021.
       doi: 10.1007/978-3-030-74432-8_6.
    """
    
    L = {v for v in G.nodes()}
    R = binary_explainable_triples(G)
    
    build = Build(R, L, mincut=mincut, weighted_mincut=weighted_mincut)
    brt = build.build_tree()
    
    return _finalize(brt, G)


# --------------------------------------------------------------------------
#                           Augmented tree
# --------------------------------------------------------------------------

def augment_and_label(tree, inplace=False):
    """Augment tree and add event labeling based on color intersections.
    
    Parameters
    ----------
    tree : Tree
        A tree whose nodes have the 'reconc' attribute.
    inplace : bool, optional
        If True, the supplied tree instance is directly manipulated, otherwise 
        the tree is copied first. The default is False.
        
    Returns
    -------
    Tree
        The unique augmented and duplication/loss-labeled tree as defined in
        [1].
    
    References
    ----------
    .. [1] D. Schaller, M. Geiß, P. F. Stadler, and M. Hellmuth.
    Complete characterization of incorrect orthology assignments in best match
    graphs. 
    In: Journal of Mathematical Biology, 82:20, 2021.
    doi: 10.1007/s00285-021-01564-8.
    """
    
    if not inplace:
        tree = tree.copy()
    
    # precompute since tree will be modified
    inner_nodes = [u for u in tree.inner_nodes()]
    leaves = tree.leaf_dict()
    
    for u in inner_nodes:
        
        # should not appear except if tree is planted, otherwise ignore
        if len(u.children) == 1:
            continue
        
        C = _color_intersection_components(u, leaves)
        
        if len(C) == 1:
            u.event = 'D'
            
        else:
            u.event = 'S'
            
            for cc in C:
                if len(cc) > 1:
                    w = TreeNode(event='D')
                    u.add_child(w)
                    for v in cc:
                        u.remove_child(v)
                        w.add_child(v)
        
    return tree


def _color_intersection_components(u, leaves):
    
    aux_graph = nx.Graph()
    
    for v in u.children:
        aux_graph.add_node(v)
        
        for x in leaves[v]:
            aux_graph.add_edge(v, x.reconc)
            
    result = []
    
    for cc in nx.connected_components(aux_graph):
        
        # discard the colors in the components
        result.append( [v for v in cc if isinstance(v, TreeNode)] )
        
    return result


# --------------------------------------------------------------------------
#                 Analysis of the structure of (R)BMGs
# --------------------------------------------------------------------------

def classify_good_ugly(bmg, rbmg, fp):
    
    for x, y in fp.edges():
        fp.edges[x, y]['middle_in_good'] = False
        fp.edges[x, y]['first_in_ugly'] = False
        fp.edges[x, y]['first_in_bad'] = False
        
    color_dict = sort_by_colors(bmg)
    
    for x, y in fp.edges():
        
        # middle edges of good quartets
        for c in color_dict:
            
            if c == bmg.nodes[x]['color'] or c == bmg.nodes[y]['color']:
                continue
            
            for z1, z2 in itertools.permutations(color_dict[c], 2):
                
                if (rbmg.has_edge(z1, x) and rbmg.has_edge(z2, y) and
                    bmg.has_edge(z1, y) and not bmg.has_edge(y, z1) and
                    bmg.has_edge(z2, x) and not bmg.has_edge(x, z2)):
                    
                    fp.edges[x, y]['middle_in_good'] = True
        
        # first edges of ugly/bad quartets          
        for _ in range(2):
            
            for x2 in color_dict[bmg.nodes[x]['color']]:
                if x == x2:
                    continue
                
                for z in rbmg.neighbors(x2):
                    
                    if (bmg.nodes[z]['color'] == bmg.nodes[x]['color'] or
                        bmg.nodes[z]['color'] == bmg.nodes[y]['color']):
                        continue
                    # first edges of ugly quartets
                    if (rbmg.has_edge(y, x2) and not rbmg.has_edge(z, y) and
                        not rbmg.has_edge(z, x)):
                        fp.edges[x, y]['first_in_ugly'] = True
                    
                    # first edges of bad quartets
                    elif (not rbmg.has_edge(y, x2) and rbmg.has_edge(z, y) and
                          not bmg.has_edge(x, z)):
                        fp.edges[x, y]['first_in_bad'] = True
            
            # swap for second iteration
            x, y = y, x
    
    return fp   


def count_good_ugly(bmg, rbmg, fp):
    
    classify_good_ugly(bmg, rbmg, fp)
    
    good, ugly, both = 0, 0, 0
    
    for x, y in fp.edges():
        
        if fp.edges[x, y]['middle_in_good'] and fp.edges[x, y]['first_in_ugly']:
            good += 1
            ugly += 1
            both += 1
        elif fp.edges[x, y]['middle_in_good']:
            good += 1
        elif fp.edges[x, y]['first_in_ugly']:
            ugly += 1
    
    return good, ugly, both


def count_good_ugly_bad(bmg, rbmg, fp):
    
    classify_good_ugly(bmg, rbmg, fp)
    
    gub, gu, gb, ub, g, u, b = 0, 0, 0, 0, 0, 0, 0
    
    for x, y in fp.edges():
        
        if (fp.edges[x, y]['middle_in_good'] and
            fp.edges[x, y]['first_in_ugly'] and
            fp.edges[x, y]['first_in_bad']):
            gub += 1
            gu += 1
            gb += 1
            ub += 1
            g += 1
            u += 1
            b += 1
            
        elif fp.edges[x, y]['middle_in_good'] and fp.edges[x, y]['first_in_ugly']:
            gu += 1
            g += 1
            u += 1
        elif fp.edges[x, y]['middle_in_good'] and fp.edges[x, y]['first_in_bad']:
            gb += 1
            g += 1
            b += 1
        elif fp.edges[x, y]['first_in_ugly'] and fp.edges[x, y]['first_in_bad']:
            ub += 1
            u += 1
            b += 1
        elif fp.edges[x, y]['middle_in_good']:
            g += 1
        elif fp.edges[x, y]['first_in_ugly']:
            u += 1
        elif fp.edges[x, y]['first_in_bad']:
            b += 1
    
    return gub, gu, gb, ub, g, u, b
    

# --------------------------------------------------------------------------
#                             P4 EDITING
# --------------------------------------------------------------------------
    
def list_path(t, k, path, G, P4_list):
    for u in G[path[-1]]:
        G.nodes[u]['vis'] += 1
    for u in G[path[-1]]:
        if G.nodes[u]['vis'] == 1:
            if t < k:
                list_path(t+1, k, path + [u], G, P4_list)
            elif path[0] < u:
                P4_list.append(tuple(path) + (u,))
    for u in G[path[-1]]:
        G.nodes[u]['vis'] -= 1


def find_all_P4(G):
    P4_list = []
    for v in G:
            G.nodes[v]['vis'] = 0
    for v in G:
        G.nodes[v]['vis'] = 1
        list_path(2, 4, [v], G, P4_list)
        G.nodes[v]['vis'] = 0
    return P4_list


def is_good_quartet(path, bmg):
    """Determine whether an induced P4 is a good quartet in the bmg."""
    if (bmg.has_edge(path[0], path[2]) and 
        bmg.has_edge(path[3], path[1]) and
        bmg.nodes[path[0]]['color'] == bmg.nodes[path[3]]['color']):
        return True
    else:
        return False
    

def remove_all_P4(rbmg, bmg, P4_list=None):
    """Find all good quartets and removes the middle edges."""
    GP4 = rbmg.copy()
    
    if P4_list is None:
        P4_list = find_all_P4(GP4)

    for path in P4_list:
        if is_good_quartet(path, bmg) and GP4.has_edge(path[1], path[2]):
            GP4.remove_edge(path[1], path[2])
            print(path[1], "-|-", path[2])
    
    return GP4

# --------------------------------------------------------------------------
#                            C6 DETECTION
# --------------------------------------------------------------------------

def validate_C6(G, C6_path):
    
    C6 = G.subgraph(C6_path)
    if C6.size() != 6:
        print(C6_path)
        print('Nodes', [n for n in C6.nodes()], C6.order())
        print('Edges', [edge for edge in C6.edges()], C6.size())
        print('Size not equal to 6!')
        return False
    for n in C6.nodes():
        if not C6.degree(n) == 2:
            print("Degree not equal to 2 for", n)
            return False
        for neighbor in G.neighbors(n):
            if G.nodes[n]['color'] == G.nodes[neighbor]['color']:
                print('Same color', n, neighbor)
                return False
    return True


def find_all_C6(G, P4_list=None):
    """Find induces C6 of the form <x1 y1 z1 x2 y2 z2>."""
    
    if P4_list is None:
        P4_list = find_all_P4(G)
        
    endpoints = {}
    for path in P4_list:
        a, b, c, d = path if path[0] < path[3] else path[::-1]
        a_col, b_col, c_col, d_col = (G.nodes[a]['color'], G.nodes[b]['color'],
                                      G.nodes[c]['color'], G.nodes[d]['color'])
        # avoid multiple recording for the 3 different colors:
        if a_col == d_col and a_col < b_col and a_col < c_col:
            if (a, d) not in endpoints:
                endpoints[(a, d)] = {}
            if (b_col, c_col) not in endpoints[(a, d)]:
                endpoints[(a, d)][(b_col, c_col)] = []
            endpoints[(a, d)][(b_col, c_col)].append(path)
    
    C6_list = []
    for endpoint, colors in endpoints.items():
        x1, x2 = endpoint
        for color_pair, paths in colors.items():
            y_color, z_color = color_pair
            # avoid multiple recording for the 2 directions:
            if y_color < z_color and (z_color, y_color) in colors:
                for path1 in paths:
                    for path2 in colors[(z_color, y_color)]:
                        if ((not G.has_edge(path1[1], path2[1])) and
                            (not G.has_edge(path1[2], path2[2]))):
                            C6_list.append(path1 + (path2[2], path2[1]))
                            if not validate_C6(G, C6_list[-1]):
                                raise RuntimeError('problem in C6 detection')
                            print(C6_list[-1])
    return C6_list


def graph_type(G):
    
    P4_list = find_all_P4(G)
    C6_list = find_all_C6(G, P4_list=P4_list)
    
    if C6_list:
        return 'C', P4_list, C6_list
    elif P4_list:
        return 'B', P4_list, C6_list
    else:
        return 'A', P4_list, C6_list
